package lcsb.mapviewer.web.api;

import lcsb.mapviewer.model.ProjectStatus;
import org.apache.commons.collections4.ListUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.restdocs.payload.FieldDescriptor;
import org.springframework.restdocs.payload.JsonFieldType;
import org.springframework.restdocs.payload.RequestFieldsSnippet;
import org.springframework.restdocs.request.RequestParametersSnippet;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import static org.springframework.restdocs.payload.PayloadDocumentation.fieldWithPath;
import static org.springframework.restdocs.payload.PayloadDocumentation.requestFields;
import static org.springframework.restdocs.payload.PayloadDocumentation.subsectionWithPath;
import static org.springframework.restdocs.request.RequestDocumentation.parameterWithName;
import static org.springframework.restdocs.request.RequestDocumentation.requestParameters;

public class NewApiDocs {

  public static List<FieldDescriptor> getBioEntityResponse(final String prefix) {
    return new NewApiDocs().getBioEntityFields(prefix);
  }

  public static List<FieldDescriptor> getElementResponse(final String prefix) {
    return new NewApiDocs().getElementFields(prefix);
  }

  public static List<FieldDescriptor> getReactionResponse(final String prefix) {
    return new NewApiDocs().getReactionFields(prefix);
  }

  public static List<FieldDescriptor> getChemicalListResponse(final String prefix) {
    return new NewApiDocs().getChemicalsFields(prefix + "[].");
  }

  public static List<FieldDescriptor> getDrugListResponse(final String prefix) {
    return new NewApiDocs().getDrugsFields(prefix + "[].");
  }

  public static List<FieldDescriptor> getMirnaListResponse(final String prefix) {
    return new NewApiDocs().getMirnaFields(prefix + "[].");
  }

  public static List<FieldDescriptor> getProjectResponse(final String prefix) {
    return new NewApiDocs().getProjectFields(prefix);
  }

  public static RequestParametersSnippet getChemicalFilter() {
    return requestParameters(
        parameterWithName("query")
            .description("name of chemical that we are searching for")
            .optional(),
        parameterWithName("target")
            .description("target element that we are searching for in format TYPE:ID")
            .optional());
  }

  public static RequestParametersSnippet getDrugFilter() {
    return requestParameters(
        parameterWithName("query")
            .description("name of drug that we are searching for")
            .optional(),
        parameterWithName("target")
            .description("target element that we are searching for in format TYPE:ID")
            .optional());
  }

  public static RequestParametersSnippet getMirnaFilter() {
    return requestParameters(
        parameterWithName("query")
            .description("name of MiRNA that we are searching for")
            .optional(),
        parameterWithName("target")
            .description("target element that we are searching for in format TYPE:ID")
            .optional());
  }

  public static RequestFieldsSnippet getReactionCsvRequest() {
    return requestFields(new NewApiDocs().getReactionCsvFields());
  }

  public static RequestFieldsSnippet getElementCsvRequest() {
    return requestFields(new NewApiDocs().getElementCsvFields());
  }

  public static List<FieldDescriptor> getOverlayBioEntityList(final String prefix) {
    return new NewApiDocs().getOverlayBioEntityPairFields(prefix + "[].");
  }

  public static List<FieldDescriptor> getPageableFields() {
    return Arrays.asList(
        fieldWithPath("totalPages")
            .description("total number of pages"),
        fieldWithPath("totalElements")
            .description("total number of elements"),
        fieldWithPath("numberOfElements")
            .description("number of elements on this page"),
        fieldWithPath("size")
            .description("page size"),
        fieldWithPath("number")
            .description("page number"));
  }


  private List<FieldDescriptor> getOverlayBioEntityPairFields(final String prefix) {
    final List<FieldDescriptor> result = new ArrayList<>();
    result.addAll(Arrays.asList(
        fieldWithPath(prefix + "left")
            .description("bioEntity")
            .type(JsonFieldType.OBJECT),
        fieldWithPath(prefix + "right")
            .description("overlay entry matching bioEntity")
            .type(JsonFieldType.OBJECT)));
    result.addAll(getBioEntityFields(prefix + "left."));
    result.addAll(getOverlayEntryFields(prefix + "right."));
    return result;
  }

  private List<FieldDescriptor> getOverlayEntryFields(final String prefix) {
    final List<FieldDescriptor> result = new ArrayList<>();
    result.addAll(Arrays.asList(
        fieldWithPath(prefix + "id")
            .description("identifier")
            .type(JsonFieldType.NUMBER),
        fieldWithPath(prefix + "name")
            .description("bioEntity name that should match")
            .type(JsonFieldType.STRING),
        fieldWithPath(prefix + "modelName")
            .description("bioEntity model name that should match")
            .optional()
            .type(JsonFieldType.STRING),
        fieldWithPath(prefix + "elementId")
            .description("bioEntity identifier from external resource (like CellDesigner file that was used for generating the map)")
            .type(JsonFieldType.STRING),
        fieldWithPath(prefix + "reverseReaction")
            .description("should the reaction be reversed in the overlay")
            .optional()
            .type(JsonFieldType.BOOLEAN),
        fieldWithPath(prefix + "lineWidth")
            .description("line width used when drawing overlay entry")
            .optional()
            .type(JsonFieldType.NUMBER),
        fieldWithPath(prefix + "value")
            .description("normalized value <-1,1> assigned to this entry")
            .optional()
            .type(JsonFieldType.NUMBER),
        fieldWithPath(prefix + "color")
            .description("color that should be used when drawing overlay entry")
            .optional()
            .type(JsonFieldType.OBJECT),
        fieldWithPath(prefix + "description")
            .description("description")
            .type(JsonFieldType.STRING)
            .optional()

    ));
    result.addAll(getColorFields(prefix + "color."));
    return result;
  }

  private List<FieldDescriptor> getReactionCsvFields() {
    return new ArrayList<>(Arrays.asList(
        fieldWithPath("elementTypes")
            .description("list of element types")
            .type(JsonFieldType.ARRAY),
        fieldWithPath("reactionTypes")
            .description("list of reaction types")
            .type(JsonFieldType.ARRAY),
        fieldWithPath("submaps")
            .description("list of map identifiers")
            .type(JsonFieldType.ARRAY),
        fieldWithPath("includedCompartmentIds")
            .description("list of compartment identifiers where at least one participant should be present")
            .type(JsonFieldType.ARRAY),
        fieldWithPath("columns")
            .description("list of columns that should be returned in csv")
            .type(JsonFieldType.ARRAY),
        fieldWithPath("annotations")
            .description("list of annotations that should be returned in csv")
            .type(JsonFieldType.ARRAY),
        fieldWithPath("excludedCompartmentIds")
            .description("list of compartment identifiers where all participant should not be present")
            .type(JsonFieldType.ARRAY)));
  }

  private List<FieldDescriptor> getElementCsvFields() {
    return new ArrayList<>(Arrays.asList(
        fieldWithPath("types")
            .description("list of element types")
            .type(JsonFieldType.ARRAY),
        fieldWithPath("submaps")
            .description("list of map identifiers")
            .type(JsonFieldType.ARRAY),
        fieldWithPath("includedCompartmentIds")
            .description("list of compartment identifiers where elements should be included")
            .type(JsonFieldType.ARRAY),
        fieldWithPath("columns")
            .description("list of columns that should be returned in csv")
            .type(JsonFieldType.ARRAY),
        fieldWithPath("annotations")
            .description("list of annotations that should be returned in csv")
            .type(JsonFieldType.ARRAY),
        fieldWithPath("excludedCompartmentIds")
            .description("list of compartment identifiers from which elements should be excluded")
            .type(JsonFieldType.ARRAY)));
  }

  private List<FieldDescriptor> getBioEntityFields(final String prefix) {
    final List<FieldDescriptor> result = getElementFields(prefix);
    result.addAll(getReactionSpecificFields(prefix));

    return result;
  }

  private List<FieldDescriptor> getElementFields(final String prefix) {
    final List<FieldDescriptor> result = getBioEntitySpecificFields(prefix);
    result.addAll(getSpeciesSpecificFields(prefix));
    result.addAll(getChemicalSpecificFields(prefix));
    result.addAll(getCompartmentSpecificFields(prefix));
    return result;
  }

  private List<FieldDescriptor> getReactionFields(final String prefix) {
    final List<FieldDescriptor> result = getBioEntitySpecificFields(prefix);
    result.addAll(getReactionSpecificFields(prefix));
    return result;
  }

  private List<FieldDescriptor> getBioEntitySpecificFields(final String prefix) {
    final List<FieldDescriptor> result = new ArrayList<>();
    result.add(
        fieldWithPath(prefix + "immediateLink")
            .description("link that should be opened immediately when element is clicked on the map")
            .optional()
            .type(JsonFieldType.STRING));
    result.add(
        fieldWithPath(prefix + "id")
            .description("unique bioEntity identifier")
            .type(JsonFieldType.NUMBER));
    result.add(
        fieldWithPath(prefix + "stringType")
            .description("type of the bioEntity")
            .type(JsonFieldType.STRING));
    result.add(

        fieldWithPath(prefix + "name")
            .description("name")
            .type(JsonFieldType.STRING)
            .optional());
    result.add(

        fieldWithPath(prefix + "elementId")
            .description("element identifier taken from source file")
            .type(JsonFieldType.STRING));
    result.add(

        fieldWithPath(prefix + "model")
            .description("map identifier")
            .type(JsonFieldType.NUMBER)

    );
    result.add(subsectionWithPath(prefix + "references")
        .description("list of references")
        .type(JsonFieldType.ARRAY)

    );

    result.add(fieldWithPath(prefix + "z")
        .description("z index (coordinate) of the element on the map")
        .type(JsonFieldType.NUMBER));

    result.add(fieldWithPath(prefix + "notes")
        .description("notes and description")
        .type(JsonFieldType.STRING));

    result.add(fieldWithPath(prefix + "symbol")
        .description("symbol")
        .type(JsonFieldType.STRING)
        .optional());

    result.add(fieldWithPath(prefix + "abbreviation")
        .description("abbreviation")
        .type(JsonFieldType.STRING)
        .optional());

    result.add(fieldWithPath(prefix + "formula")
        .description("formula")
        .type(JsonFieldType.STRING)
        .optional());

    result.add(subsectionWithPath(prefix + "synonyms")
        .description("list of synonyms")
        .type(JsonFieldType.ARRAY)
        .optional());

    result.add(fieldWithPath(prefix + "visibilityLevel")
        .description("at what zoom level this element becomes visible in hierarchical view")
        .type(JsonFieldType.STRING)
        .optional());

    result.add(fieldWithPath(prefix + "homodimer")
        .description("multimer value")
        .type(JsonFieldType.NUMBER)
        .optional());
    return result;
  }

  private List<FieldDescriptor> getSpeciesSpecificFields(final String prefix) {
    final List<FieldDescriptor> result = new ArrayList<>(Arrays.asList(
        fieldWithPath(prefix + "nameX")
            .description("x coordinate of the element name on the map")
            .type(JsonFieldType.NUMBER)
            .optional(),

        fieldWithPath(prefix + "nameY")
            .description("y coordinate of the element name on the map")
            .type(JsonFieldType.NUMBER)
            .optional(),

        fieldWithPath(prefix + "nameWidth")
            .description("width of the box with element name")
            .type(JsonFieldType.NUMBER)
            .optional(),

        fieldWithPath(prefix + "nameHeight")
            .description("height of the box with element name")
            .type(JsonFieldType.NUMBER)
            .optional(),

        fieldWithPath(prefix + "nameVerticalAlign")
            .description("vertical align of the text inside the box with element name")
            .type(JsonFieldType.STRING)
            .optional(),

        fieldWithPath(prefix + "nameHorizontalAlign")
            .description("horizontal align of the text inside the box with element name")
            .type(JsonFieldType.STRING)
            .optional(),

        fieldWithPath(prefix + "width")
            .description("width of the element")
            .type(JsonFieldType.NUMBER)
            .optional(),

        fieldWithPath(prefix + "height")
            .description("height of the element")
            .type(JsonFieldType.NUMBER)
            .optional(),

        fieldWithPath(prefix + "transparencyLevel")
            .description("at what zoom level this element becomes transparent in hierarchical view")
            .type(JsonFieldType.STRING)
            .optional(),

        subsectionWithPath(prefix + "formerSymbols")
            .description("list of former symbols")
            .type(JsonFieldType.ARRAY)
            .optional(),

        fieldWithPath(prefix + "fullName")
            .description("full name")
            .type(JsonFieldType.STRING)
            .optional(),

        fieldWithPath(prefix + "glyph")
            .description("image glyph associated with the element")
            .type(JsonFieldType.OBJECT)
            .optional(),

        fieldWithPath(prefix + "glyph.id")
            .description("glyph id")
            .type(JsonFieldType.NUMBER)
            .optional(),

        fieldWithPath(prefix + "glyph.file")
            .description("id of file associated with the glyph")
            .type(JsonFieldType.NUMBER)
            .optional(),

        fieldWithPath(prefix + "activity")
            .description("is the element active")
            .type(JsonFieldType.BOOLEAN)
            .optional(),

        fieldWithPath(prefix + "hypothetical")
            .description("is the element hypothetical")
            .type(JsonFieldType.BOOLEAN)
            .optional(),

        fieldWithPath(prefix + "boundaryCondition")
            .description("SBML kinetics is boundary condition")
            .type(JsonFieldType.BOOLEAN)
            .optional(),

        fieldWithPath(prefix + "constant")
            .description("SBML kinetics is constant")
            .type(JsonFieldType.BOOLEAN)
            .optional(),

        fieldWithPath(prefix + "initialAmount")
            .description("SBML kinetics initial amount")
            .type(JsonFieldType.NUMBER)
            .optional(),

        fieldWithPath(prefix + "initialConcentration")
            .description("SBML kinetics initial concentration")
            .type(JsonFieldType.NUMBER)
            .optional(),

        fieldWithPath(prefix + "charge")
            .description("SBML charge")
            .type(JsonFieldType.NUMBER)
            .optional(),

        fieldWithPath(prefix + "substanceUnits")
            .description("SBML substance units")
            .type(JsonFieldType.STRING)
            .optional(),

        fieldWithPath(prefix + "onlySubstanceUnits")
            .description("SBML only substance units")
            .type(JsonFieldType.BOOLEAN)
            .optional(),

        subsectionWithPath(prefix + "modificationResidues")
            .description("List of species modification residues")
            .type(JsonFieldType.ARRAY)
            .optional(),

        subsectionWithPath(prefix + "structuralState")
            .description("Species structural state")
            .type(JsonFieldType.OBJECT)
            .optional(),

        fieldWithPath(prefix + "complex")
            .description("identifier of a complex in which element is located")
            .type(JsonFieldType.NUMBER)
            .optional(),

        fieldWithPath(prefix + "complexName")
            .description("name of a complex in which element is located")
            .type(JsonFieldType.STRING)
            .optional(),

        fieldWithPath(prefix + "compartment")
            .description("identifier of a compartment in which element is located")
            .type(JsonFieldType.NUMBER)
            .optional(),

        fieldWithPath(prefix + "pathway")
            .description("identifier of a pathway in which element is located")
            .type(JsonFieldType.NUMBER)
            .optional(),

        fieldWithPath(prefix + "compartmentName")
            .description("name of a compartment in which element is located")
            .type(JsonFieldType.STRING)
            .optional(),

        fieldWithPath(prefix + "submodel")
            .description("submap to which this bioEntity is an entry point (anchor)")
            .type(JsonFieldType.OBJECT)
            .optional(),

        fieldWithPath(prefix + "submodel.mapId")
            .description("identifier of a submap to which this bioEntity is an entry point (anchor)")
            .type(JsonFieldType.NUMBER)
            .optional(),

        fieldWithPath(prefix + "submodel.type")
            .description("type of submap connection")
            .type(JsonFieldType.STRING)
            .optional(),

        fieldWithPath(prefix + "x")
            .description("x coordinate of the element on the map")
            .type(JsonFieldType.NUMBER)
            .optional(),

        fieldWithPath(prefix + "y")
            .description("y coordinate of the element on the map")
            .type(JsonFieldType.NUMBER)
            .optional(),
        fieldWithPath(prefix + "lineWidth")
            .description("border line width")
            .type(JsonFieldType.NUMBER)
            .optional()

    ));
    result.add(fieldWithPath(prefix + "fontColor")
        .description("font color")
        .type(JsonFieldType.OBJECT)
        .optional());

    result.addAll(getColorFields(prefix + "fontColor."));
    result.add(fieldWithPath(prefix + "fontSize")
        .description("font size")
        .type(JsonFieldType.NUMBER)
        .optional());
    result.add(fieldWithPath(prefix + "fillColor")
        .description("fill color")
        .type(JsonFieldType.OBJECT)
        .optional());
    result.addAll(getColorFields(prefix + "fillColor."));
    result.add(fieldWithPath(prefix + "borderColor")
        .description("border color")
        .optional()
        .type(JsonFieldType.OBJECT));
    result.addAll(getColorFields(prefix + "borderColor."));

    return result;
  }

  private List<FieldDescriptor> getChemicalSpecificFields(final String prefix) {
    return Arrays.asList(
        fieldWithPath(prefix + "smiles")
            .description("smiles identifier (for chemicals)")
            .type(JsonFieldType.STRING)
            .optional(),

        fieldWithPath(prefix + "inChI")
            .description("inChI identifier (for chemicals)")
            .type(JsonFieldType.STRING)
            .optional(),

        fieldWithPath(prefix + "inChIKey")
            .description("inChI key identifier (for chemicals)")
            .type(JsonFieldType.STRING)
            .optional());
  }

  private List<FieldDescriptor> getCompartmentSpecificFields(final String prefix) {
    return Arrays.asList(
        fieldWithPath(prefix + "thickness")
            .description("compartment membrane thickness")
            .type(JsonFieldType.NUMBER)
            .optional(),
        fieldWithPath(prefix + "outerWidth")
            .description("compartment outer line width")
            .type(JsonFieldType.NUMBER)
            .optional(),
        fieldWithPath(prefix + "innerWidth")
            .description("compartment inner line width")
            .type(JsonFieldType.NUMBER)
            .optional());
  }

  private List<FieldDescriptor> getReactionSpecificFields(final String prefix) {
    final List<FieldDescriptor> result = new ArrayList<>();
    result.addAll(Arrays.asList(
        fieldWithPath(prefix + "idReaction")
            .description("reaction identifier taken from source file")
            .type(JsonFieldType.STRING)
            .optional(),
        fieldWithPath(prefix + "reversible")
            .description("is reaction reversible")
            .type(JsonFieldType.BOOLEAN)
            .optional(),
        fieldWithPath(prefix + "mechanicalConfidenceScore")
            .description("reaction mechanical confidence score (from RECON)")
            .type(JsonFieldType.NUMBER)
            .optional(),
        fieldWithPath(prefix + "lowerBound")
            .description("reaction lower bound (from RECON)")
            .type(JsonFieldType.NUMBER)
            .optional(),
        fieldWithPath(prefix + "upperBound")
            .description("reaction upper bound (from RECON)")
            .type(JsonFieldType.NUMBER)
            .optional(),
        fieldWithPath(prefix + "subsystem")
            .description("reaction subsystem (from RECON)")
            .type(JsonFieldType.STRING)
            .optional(),
        fieldWithPath(prefix + "geneProteinReaction")
            .description("reaction gene protein reaction (from RECON)")
            .type(JsonFieldType.STRING)
            .optional(),
        subsectionWithPath(prefix + "kinetics")
            .description("reaction SBML kinetics law")
            .type(JsonFieldType.OBJECT)
            .optional(),

        fieldWithPath(prefix + "products")
            .description("list of reaction products")
            .type(JsonFieldType.ARRAY)
            .optional(),
        fieldWithPath(prefix + "products[].id")
            .description("identifier of the product")
            .type(JsonFieldType.NUMBER)
            .ignored(),
        fieldWithPath(prefix + "products[].element")
            .description("identifier of the element for this product")
            .type(JsonFieldType.NUMBER)
            .optional(),
        fieldWithPath(prefix + "products[].stoichiometry")
            .description("SBML stoichiometry")
            .type(JsonFieldType.NUMBER)
            .optional(),
        fieldWithPath(prefix + "products[].line")
            .description("line used to draw product connection")
            .type(JsonFieldType.OBJECT)
            .optional(),

        fieldWithPath(prefix + "reactants")
            .description("list of reaction reactants")
            .type(JsonFieldType.ARRAY)
            .optional(),
        fieldWithPath(prefix + "reactants[].id")
            .description("identifier of the reactant")
            .type(JsonFieldType.NUMBER)
            .ignored(),
        fieldWithPath(prefix + "reactants[].element")
            .description("identifier of the element for this reactant")
            .type(JsonFieldType.NUMBER)
            .optional(),
        fieldWithPath(prefix + "reactants[].stoichiometry")
            .description("SBML stoichiometry")
            .type(JsonFieldType.NUMBER)
            .optional(),
        fieldWithPath(prefix + "reactants[].line")
            .description("line used to draw reactant connection")
            .type(JsonFieldType.OBJECT)
            .optional(),

        fieldWithPath(prefix + "modifiers")
            .description("list of reaction modifiers")
            .type(JsonFieldType.ARRAY)
            .optional(),
        fieldWithPath(prefix + "modifiers[].id")
            .description("identifier of the modifier")
            .type(JsonFieldType.NUMBER)
            .ignored(),
        fieldWithPath(prefix + "modifiers[].element")
            .description("identifier of the element for this modifier")
            .type(JsonFieldType.NUMBER)
            .optional(),
        fieldWithPath(prefix + "modifiers[].stoichiometry")
            .description("SBML stoichiometry")
            .type(JsonFieldType.NUMBER)
            .optional(),
        fieldWithPath(prefix + "modifiers[].line")
            .description("line used to draw modifier connection")
            .type(JsonFieldType.OBJECT)
            .optional(),

        subsectionWithPath(prefix + "processCoordinates")
            .description(
                "coordinates of process in the reaction if defined manually. If this value is not present then center of line should be taken.")
            .type(JsonFieldType.OBJECT)
            .optional(),
        fieldWithPath(prefix + prefix + "line")
            .description("reaction central line")
            .type(JsonFieldType.OBJECT)
            .optional()));
    result.addAll(getLineFields(prefix + "line."));
    result.addAll(getLineFields(prefix + "products[].line."));
    result.addAll(getLineFields(prefix + "reactants[].line."));
    result.addAll(getLineFields(prefix + "modifiers[].line."));
    result.add(fieldWithPath(prefix + "operators")
        .description("reaction operators")
        .type(JsonFieldType.ARRAY)
        .optional());
    result.addAll(getOperatorFields(prefix + "operators[]."));
    return result;
  }

  private List<FieldDescriptor> getLineFields(final String prefix) {
    return ListUtils.union(Arrays.asList(
            fieldWithPath(prefix + "id")
                .description("line identifier")
                .type(JsonFieldType.NUMBER)
                .optional(),
            fieldWithPath(prefix + "width")
                .description("line width")
                .type(JsonFieldType.NUMBER)
                .optional(),
            fieldWithPath(prefix + "z")
                .description("line z coordinate (z-index)")
                .type(JsonFieldType.NUMBER)
                .optional(),
            subsectionWithPath(prefix + "segments")
                .description("line segments")
                .type(JsonFieldType.ARRAY)
                .optional(),
            subsectionWithPath(prefix + "startArrow")
                .description("arrow definition of the line beginning")
                .type(JsonFieldType.OBJECT)
                .optional(),
            subsectionWithPath(prefix + "endArrow")
                .description("arrow definition of the line ending")
                .type(JsonFieldType.OBJECT)
                .optional(),
            fieldWithPath(prefix + "lineType")
                .description("line pattern")
                .type(JsonFieldType.STRING)
                .optional(),
            fieldWithPath(prefix + "color")
                .description("line color")
                .type(JsonFieldType.OBJECT)
                .optional()),
        getColorFields(prefix + "color."));
  }

  private List<FieldDescriptor> getOperatorFields(final String prefix) {
    return Arrays.asList(
        fieldWithPath(prefix + "id")
            .description("operator identifier")
            .type(JsonFieldType.NUMBER)
            .optional(),
        subsectionWithPath(prefix + "line")
            .description("line for the operator")
            .type(JsonFieldType.OBJECT)
            .optional(),
        subsectionWithPath(prefix + "inputs").optional().ignored(),
        subsectionWithPath(prefix + "outputs").optional().ignored(),
        fieldWithPath(prefix + "operatorText").optional().ignored(),
        fieldWithPath(prefix + "reactantOperator").optional().ignored(),
        fieldWithPath(prefix + "productOperator").optional().ignored(),
        fieldWithPath(prefix + "modifierOperator").optional().ignored());
  }

  private List<FieldDescriptor> getColorFields(final String prefix) {
    return Arrays.asList(
        fieldWithPath(prefix + "alpha")
            .description("alpha value (0-255)")
            .type(JsonFieldType.NUMBER)
            .optional(),
        fieldWithPath(prefix + "rgb")
            .description("RED-GREEN-BLUE value as integer #RRGGBB")
            .type(JsonFieldType.NUMBER)
            .optional());
  }

  private List<FieldDescriptor> getChemicalsFields(final String prefix) {
    return Arrays.asList(
        subsectionWithPath(prefix + "id")
            .description("identifier of the chemical")
            .type(JsonFieldType.OBJECT),
        fieldWithPath(prefix + "name")
            .description("name")
            .type(JsonFieldType.STRING),
        fieldWithPath(prefix + "description")
            .description("description")
            .type(JsonFieldType.STRING),
        fieldWithPath(prefix + "synonyms")
            .description("list of synonyms")
            .type(JsonFieldType.ARRAY),
        subsectionWithPath(prefix + "references")
            .description("list of references that identify sources")
            .type(JsonFieldType.ARRAY),
        fieldWithPath(prefix + "targets")
            .description("list of targets")
            .type(JsonFieldType.ARRAY),
        fieldWithPath(prefix + "targets[].name")
            .description("target name")
            .type(JsonFieldType.STRING),
        subsectionWithPath(prefix + "targets[].references")
            .description("list of target references")
            .type(JsonFieldType.ARRAY),
        subsectionWithPath(prefix + "targets[].targetElements")
            .description("list of elements on the map associated with this target")
            .type(JsonFieldType.ARRAY),
        subsectionWithPath(prefix + "targets[].targetParticipants")
            .description("list of identifiers associated with this target")
            .type(JsonFieldType.ARRAY),
        fieldWithPath(prefix + "directEvidence")
            .description("direct evidence")
            .optional()
            .ignored()
            .type(JsonFieldType.STRING),
        subsectionWithPath(prefix + "directEvidenceReferences")
            .description("list of references")
            .ignored()
            .type(JsonFieldType.ARRAY));
  }

  private List<FieldDescriptor> getDrugsFields(final String prefix) {
    return Arrays.asList(
        subsectionWithPath(prefix + "id")
            .description("identifier of the drug")
            .type(JsonFieldType.STRING),
        fieldWithPath(prefix + "name")
            .description("name")
            .type(JsonFieldType.STRING),
        fieldWithPath(prefix + "description")
            .description("description")
            .optional()
            .type(JsonFieldType.STRING),
        fieldWithPath(prefix + "synonyms")
            .description("list of synonyms")
            .type(JsonFieldType.ARRAY),
        fieldWithPath(prefix + "brandNames")
            .description("list of brand names")
            .type(JsonFieldType.ARRAY),
        fieldWithPath(prefix + "bloodBrainBarrier")
            .description("does drug cross blood brain barrier")
            .type(JsonFieldType.STRING),
        subsectionWithPath(prefix + "references")
            .description("list of references that identify sources")
            .type(JsonFieldType.ARRAY),
        fieldWithPath(prefix + "targets")
            .description("list of targets")
            .type(JsonFieldType.ARRAY),
        fieldWithPath(prefix + "targets[].name")
            .description("target name")
            .type(JsonFieldType.STRING),
        subsectionWithPath(prefix + "targets[].references")
            .description("list of target references")
            .type(JsonFieldType.ARRAY),
        subsectionWithPath(prefix + "targets[].targetElements")
            .description("list of elements on the map associated with this target")
            .type(JsonFieldType.ARRAY),
        subsectionWithPath(prefix + "targets[].targetParticipants")
            .description("list of identifiers associated with this target")
            .type(JsonFieldType.ARRAY));
  }

  private List<FieldDescriptor> getMirnaFields(final String prefix) {
    return Arrays.asList(
        subsectionWithPath(prefix + "id")
            .description("identifier of the MiRNA")
            .type(JsonFieldType.STRING),
        fieldWithPath(prefix + "name")
            .description("name")
            .type(JsonFieldType.STRING),
        fieldWithPath(prefix + "targets")
            .description("list of targets")
            .type(JsonFieldType.ARRAY),
        fieldWithPath(prefix + "targets[].name")
            .description("target name")
            .type(JsonFieldType.STRING),
        subsectionWithPath(prefix + "references")
            .description("list of references that identify sources")
            .type(JsonFieldType.ARRAY),
        subsectionWithPath(prefix + "targets[].references")
            .description("list of target references")
            .type(JsonFieldType.ARRAY),
        subsectionWithPath(prefix + "targets[].targetElements")
            .description("list of elements on the map associated with this target")
            .type(JsonFieldType.ARRAY),
        subsectionWithPath(prefix + "targets[].targetParticipants")
            .description("list of identifiers associated with this target")
            .type(JsonFieldType.ARRAY));
  }

  public List<FieldDescriptor> getProjectFields(final String prefix) {
    return Arrays.asList(
        fieldWithPath(prefix + "projectId")
            .description("project identifier")
            .type(JsonFieldType.STRING),
        fieldWithPath(prefix + "name")
            .description("name")
            .type(JsonFieldType.STRING),
        fieldWithPath(prefix + "sharedInMinervaNet")
            .description("is the project shared in minerva net")
            .type(JsonFieldType.BOOLEAN),
        fieldWithPath(prefix + "version")
            .description("version")
            .type(JsonFieldType.STRING),
        fieldWithPath(prefix + "owner.login")
            .description("login of the project creator")
            .type(JsonFieldType.STRING),
        fieldWithPath(prefix + "creationDate")
            .description("when project was uploaded")
            .type(JsonFieldType.STRING),
        subsectionWithPath(prefix + "disease")
            .description("identifier of the disease")
            .type("Reference")
            .optional(),
        fieldWithPath(prefix + "diseaseName")
            .description("disease name")
            .type(JsonFieldType.STRING)
            .optional(),
        subsectionWithPath(prefix + "organism")
            .description("identifier of the organism")
            .type("Reference")
            .optional(),
        fieldWithPath(prefix + "organismName")
            .description("organism name")
            .type(JsonFieldType.STRING)
            .optional(),
        fieldWithPath(prefix + "id")
            .type(JsonFieldType.NUMBER)
            .ignored(),
        fieldWithPath(prefix + "directory")
            .description(
                "directory where files related to the project are located. "
                    + "Whole path looks like: /minerva/map_images/{directory}")
            .type(JsonFieldType.STRING),
        fieldWithPath(prefix + "status")
            .description(
                "status of the uploaded project; possible values: " + getOptionsAsString(ProjectStatus.class))
            .type(JsonFieldType.STRING),
        fieldWithPath(prefix + "progress")
            .description("how much current stage progressed in percent")
            .type(JsonFieldType.NUMBER),
        fieldWithPath(prefix + "notifyEmail")
            .description("email address connected to the project")
            .type(JsonFieldType.STRING),
        fieldWithPath(prefix + "license")
            .description("project license")
            .optional()
            .type(JsonFieldType.OBJECT),
        fieldWithPath(prefix + "license.id")
            .description("project license id")
            .optional()
            .type(JsonFieldType.NUMBER),
        fieldWithPath(prefix + "license.name")
            .description("project license name")
            .optional()
            .type(JsonFieldType.STRING),
        fieldWithPath(prefix + "license.url")
            .description("project license url")
            .optional()
            .type(JsonFieldType.STRING),
        fieldWithPath(prefix + "license.content")
            .description("project license content")
            .optional()
            .type(JsonFieldType.STRING),
        fieldWithPath(prefix + "customLicenseName")
            .description("if custom license is defined here is the license name")
            .type(JsonFieldType.STRING),
        fieldWithPath(prefix + "customLicenseUrl")
            .description("if custom license is defined here is the license url")
            .type(JsonFieldType.STRING),
        fieldWithPath(prefix + "topMap.id")
            .description("id of the top map")
            .type(JsonFieldType.NUMBER),
        fieldWithPath(prefix + "logEntries")
            .description("flag indicating that there are log entries attached to the project")
            .type(JsonFieldType.BOOLEAN),
        subsectionWithPath(prefix + "overviewImageViews")
            .description("list of overview images")
            .type("array<OverviewImage>")
            .optional(),
        subsectionWithPath(prefix + "topOverviewImage")
            .description("overview image that should be used as top level image")
            .type("OverviewImage")
            .optional());
  }

  private <T extends Enum<T>> String getOptionsAsString(final Class<T> enumType) {
    final List<String> options = new ArrayList<>();
    for (final T c : enumType.getEnumConstants()) {
      options.add(c.name());
    }
    Collections.sort(options);
    return StringUtils.join(options, ", ");
  }
}
